
import 'package:logger_util/src/printers/color_printer.dart';

import 'filters/development_filter.dart';
import 'log_filter.dart';
import 'log_output.dart';
import 'log_printer.dart';
import 'outputs/console_output.dart';

/// [Level]s to control logging output. Logging can be enabled to include all
/// levels above certain [Level].
enum Level {
  verbose,
  debug,
  info,
  warning,
  error,
}

class LogEvent {
  final Level level;
  final dynamic message;
  final StackTrace? stackTrace;

  LogEvent(this.level, this.message, this.stackTrace);
}

class OutputEvent {
  final Level level;
  final bool isSave;
  final String savePath;
  final List<String> lines;

  OutputEvent(this.level, this.isSave, this.savePath, this.lines);
}

/// Use instances of logger to send log messages to the [LogPrinter].
class Logger {
  /// The current logging level of the app.
  ///
  /// All logs with levels below this level will be omitted.
  static Level level = Level.verbose;

  final LogFilter _filter;
  final LogPrinter _printer;
  final LogOutput _output;
  final bool _isSave;
  final String _savePath;
  bool _active = true;

  /// Create a new instance of Logger.
  ///
  /// You can provide a custom [printer], [filter] and [output]. Otherwise the
  /// defaults: [PrettyPrinter], [DevelopmentFilter] and [ConsoleOutput] will be
  /// used.
  Logger({
    LogFilter? filter,
    LogPrinter? printer,
    LogOutput? output,
    Level? level,
    bool isSave = false,
    String savePath = '',
  })  : _filter = filter ?? DevelopmentFilter(),
        _printer = printer ?? ColorPrinter(),
        _output = output ?? ConsoleOutput(),
        _isSave = isSave,
        _savePath = savePath {
    _filter.init();

    _filter.level = level ?? Logger.level;

    _printer.init();

    _output.init();
  }

  /// Log a message at level [Level.verbose].
  void v(dynamic message, [StackTrace? stackTrace]) {
    log(Level.verbose, message, stackTrace);
  }

  /// Log a message at level [Level.debug].
  void d(dynamic message, [StackTrace? stackTrace]) {
    log(Level.debug, message, stackTrace);
  }

  /// Log a message at level [Level.info].
  void i(dynamic message, [StackTrace? stackTrace]) {
    log(Level.info, message, stackTrace);
  }

  /// Log a message at level [Level.warning].
  void w(dynamic message, [StackTrace? stackTrace]) {
    log(Level.warning, message, stackTrace);
  }

  /// Log a message at level [Level.error].
  void e(dynamic message, [StackTrace? stackTrace]) {
    log(Level.error, message, stackTrace);
  }

  /// Log a message with [level].
  void log(Level level, dynamic message, [StackTrace? stackTrace]) {
    if (!_active) {
      throw ArgumentError('Logger has already been closed.');
    }
    var logEvent = LogEvent(level, message, stackTrace);
    if (_filter.shouldLog(logEvent)) {
      var output = _printer.log(logEvent);

      if (output.isNotEmpty) {
        var outputEvent = OutputEvent(level, _isSave, _savePath, output);
        // Issues with log output should NOT influence
        // the main software behavior.
        try {
          _output.output(outputEvent);
        } catch (e, s) {
          print(e);
          print(s);
        }
      }
    }
  }

  /// Closes the logger and releases all resources.
  void close() {
    _active = false;
    _filter.destroy();
    _printer.destroy();
    _output.destroy();
  }
}
